/*

Sid Spanners 3: The Nuthouse

Version 1.2 - 14th April 2011

Digital Prawn 2011

This program is in the public domain.

All code and data may be freely reused.

History:
10th January 2011 - Work is started on Sid Spanners 3, although after a couple of days I get bored of it and watch the entire
 run of Babylon 5 instead

21st March 2011 - After a two month break, work resumes on the game again and the pre-release version is completed and
 released to interested forum members

26th March 2011 - The fully completed version 1.0 of the game is publicly released on the WoS archive

1st April 2011 - Playing in the online Java emulator, a workmate discovers an obscure bug in the collision detection routine,
 but at the time, I dismiss it as an insignificant glitch

3rd April 2011 - Pavero releases his map of SS3. The map shows three places where inter-room doors are slightly out of
 alignment

4th April 2011 - I change my mind about SS3 being the final game in the series and with some fresh ideas, start work on
 the following sequel, Sid Spanners 4: Timebounce

7th April 2011 - SS4 is derived mainly from SS3 code. Whilst playtesting the tenth room of SS4, I manage to reproduce the bug
 found earlier by my workmate in SS3. The bug allows the player character to diagonally fall or jump right through an enemy
 sprite near the bottom of the play area without losing a life. I fix the bug in SS4. Whilst I'm at it, I also fix the
 "Conveyor bug" in SS4.

9th April 2011 - The pre-release version of SS4 is released to various forum members for testing and the game is just waiting for addition of an intro tune, loading screen and instructions file.

10th April 2011 - Meanwhile, I decide to "backport" the bugfixes from SS4 into SS3 and also fix the door misalignments with a view
 to releasing v1.1 of Sid Spanners 3

11th April 2011 - After v1.1 is uploaded to the WoS archive, I discover two further bugs, leading me to start work on v1.2.

14th April 2011 - Fixed the "Nellie Nutjob death bug". Fix the "Intro screen auto exit bug" by removing some Kempston
detection code from the music player.

Memory Map

23296 First byte after screen (Printer buffer, system variables, CHANNEL info)	} Can be used by heap provided IM2 mode in effect
23755 Start of BASIC loader							} This area can be used by the heap regardless

23296 start of malloc() heap in this program

[24576 ($6000)] Compiled C code and data should start here or above (such that BASIC loader
does not get corrupted before RANDOMIZE USR call to C program)(Start off with 27000 until size of heap is refined/reduced)

24578 - Actual start of C code and data in this program (At end of malloc() heap)
(Max size of C code/data block 24574 bytes, in this version it is a bit over 17Kbytes, so 7K of space free for possible future expansions)

49152 - start of Mister Beep's tune (2476 bytes)

51628 - start of spare 1108-byte block after Mister Beep's tune, before the game font

52736 ($CE00) 512-byte custom font, loaded in by loader (Uppercase only - chars 32-95 inclusive)
53248 ($D000) Start of IM2 vector table
53505 ($D101) Lowest possible byte used by stack (Allocated size=208 bytes)
53711 Highest word (address wise) at the base of the stack (stack base set to 53711)
53713 ($D1D1) 3-byte JP into ISR routine (Intro Music player) or 2-byte RETI instruction (Main game)
53716-53740 ($D1D4-$D1EC) 25 bytes free (Used for "My system variables" in LIEU of usual sysvars)
53741-65535 ($D1ED-$FFFF) top 12K of RAM reserved for use by SP1 sprite engine

*/

//The Makefile used to build this program follows in the below comment.
//This source file builds only the progam file, within the main release tapfile.
//To build the entire tape, including the associated datafiles referenced below,
//the UNIX 'cat' command is used to concatenate all tapfiles together.
//The other tapfiles referenced in the below Makefile can be extracted using something
//like ZX-Blockeditor from the main release tapfile.

/* 
sidspan3.tap: loader.tap screen.tap mrbeep.tap font.tap ss3code.tap
	cat loader.tap screen.tap mrbeep.tap font.tap ss3code.tap > sidspan3.tap

ss3code3.tap: sidspan3.c
	zcc +zx -osidspan3.bin -lndos -lmalloc -lim2 -lsp1 sidspan3.c -zorg=24578
	bin2tap -n sidspan3.bin ss3code.tap 24578
*/

/* List of constituent tapefiles

loader.tap: Simple BASIC loader, created using EightyOne emulator.
screen.tap: Loading screen exported using the Windows app "ZX-Paintbrush"
music.tap: Music player and data file exported using the Windows app "Beepola"
font.tap: Uppercase font file, exported using "Cygnus Font Editor" for ZX-Spectrum
sidspan3.tap: Created by Z88DK and a modified bin2tap from compiling this source file

I modified my own copy of bin2tap to not create a (superfluous) BASIC loader if the "-n"
flag is specified (source code of the modfied program on WoS forums). 
Although, bin2code by metalbrain  ( http://www.speccy.org/metalbrain/bin2code.c ) can be used instead, 
to do the same thing.
*/

//Includes
#include <sprites/sp1.h>
#include <malloc.h>
#include <spectrum.h>

#include <stdlib.h>
#include <string.h>

//Only needed for debugging
//This game doesn't use stdio.h functions
//In "normal" usage don't include it to save memory
//#include <stdio.h>

#include <input.h>
#include <im2.h>

//Preprocessor directives
#pragma output STACKPTR=53713	// Set stack at 208-byte gap between IM2 vector table and ISR entry point
#define SGN(x) ((x)<0?-1:((x)?1:0))
#define ABS(x) ((x)<0?-(x):(x))
//Macro to get/set screen attribute
#define MY_ATTR(row,col) *(byte *)(0x5800+((row)<<5)+(col))
//May need to define NULL if stdio.h not included
#ifndef NULL
#define NULL (void *)0
#endif

//#define WAIT_KEY while(!in_Inkey())
#define WAIT_KEY in_WaitForKey()
#define FONT_ADDR 52736	//Address of custom font
#define BEEPOLA_ADDR 50688 //Address of player & tune exported from Beepola
#define MY_BORDCR 53716 //Address of MY_BORDCR "pseudo-system variable"
#define INITIAL_ROOM 1 //Starting room
#define HEAP_ADDR 23296
#define HEAP_SIZE 1282  //Exact size of heap needed for the in-game sprites
#define NUM_SPRITES 5
#define X_OFFSET 88 //"eXtra Data Offset" Each room has 88 bytes of map data followed by "Extra data". This constant
			// allows convenient access to the extra data elements of the room data array.
#define O_OFFSET 110 //Offset of object data in room data array

//Game parameters
#define START_ROOM 1
//Version 1.1 - Number of lives reduced back to 8 as in the pre-release version (16 in v1.0 was too many)
#define NUM_LIVES 8
//#define NUM_LIVES 16
#define INVULNERABLE_TIME 100
#define MOVEPLAT_DELAY 10
#define STARTING_ROW 19
#define STARTING_COL 29

//Typedefs
typedef unsigned char byte;
typedef unsigned int word;

//Sprite data structure
struct sprentry
	{
	struct sp1_ss *s;
	char dx;
	char dy;
	byte frame;
	//pixel position of sprite on screen (used for collision detection and bounding)
    	byte px;
	byte py;
	//start positions (player and guardians)
	byte startcol;
	byte startrow;
	//end positions (guardians only)
	//For the player character these two bytes instead specify 
	//the starting hrot and vrot parameters
	byte endcol;
	byte endrow;
	//for guardians, direction: 1=travelling towards end, -1=travelling back to start
	char direction;
	};

//Global variables
//heap pointer for malloc()
long heap;
//Function pointer to input routine
void *getjoy;
//Structure for keyboard input
struct in_UDK keys;
//current room number, flag set when room is exited
byte room,room_exited;
//Pointer to current room data
byte *room_data;
//Global variable for passing into "hook functions"
byte global_n;
//Number of lives
byte lives;
//Number of nuts in the entire game
word nuts;
//Number of guardians in current room
byte guardians;
//Destination room and direction travelling
byte dest_room,travel_dir;
//Main Game loop variables and timers
byte airborne, facing_left, frame_tick;
byte jump_tick, guardian_flip, guardian_tick;
byte solid_ground,airborne_entry;
byte invulnerable;
//Conveyor belt and moving platform variables
byte conveyor,convrow,convpos;
byte moveplat,mprow,mpstartcol,mpendcol,mpcol,mpleft,mptick;
//Sprite Collision detection
byte caught,cx,cy;
//Tile detection
word touchloc;
byte tx,ty,my_attr;
char touched;
//keyboard input byte
byte input;
//Pointer to sprite entry being processed
struct sprentry *se;
//Sprite colours (Varies in each room)
byte sprite_colour[NUM_SPRITES];
//Global text cursor position
byte textrow,textcol;
//Temp buffer for status area and debugging messages
//(one line long plus an extra byte for a NUL terminator)
char tempstr[33];
//The sprite table
struct sprentry sprtbl[NUM_SPRITES];
//Array of pointers to animated sequence of frames of Sid character walking
unsigned char *anim_sid_left[6];
unsigned char *anim_sid_right[6];

//Array of pointers to "Start frames" of each sprite
//Next frame of guardian sprites is guaranteed to be 48 bytes 
//later in memory that initial frame
unsigned char *start_frame[NUM_SPRITES];

//Define play-area clipping rectangle used by SP1 library
struct sp1_Rect cr = {0, 0, 32, 22};		//Top twenty-two lines of the screen
//Smaller rectangle for invalidating 1x1 character square
struct sp1_Rect onecr = {0,0,1,1};
//Clipping rectangle for conveyor belt
struct sp1_Rect conveyor_cr = {0,4,24,1};
//Clipping rectangle for moving platform
struct sp1_Rect moveplat_cr = {0,1,30,1};
//Nellie nutjob clipping rectangle
struct sp1_Rect nellie_cr = {19,7,2,2};

//Allow fast selection of bits within byte
byte bit_selector[8]={0x80,0x40,0x20,0x10,0x08,0x04,0x02,0x01};

//Colour of bricks on current level
//"Ordinary" bricks have this colour
//Bricks around the edges have this colour or'd with "mortar_colour" paper
//This is needed for ATTR detection of platforms
byte brick_colour, mortar_colour;

/*Colour scheme  (unless specified otherwise, all colours are BRIGHT)
0 BLACK unused

1 BLUE        These four can be used by guardian sprites and brick colours too
2 RED      
3 MAGENTA
4 GREEN

5 CYAN - reserved for possible additional in-game objects

6 YELLOW - Sid spanners sprite and empty spaces have yellow ink
7 WHITE - nuts
*/

//Extra data format
//(&extradata[0]=&room_data[X_OFFSET])
//
//extradata[0]
//conveyor belt
//bit 7: 1=moving left, 0=moving right
//bits 6,5,4 - columns processed by conveyor belt
//bit 6= columns 4-11, bit 5= columns 12-19, bit 4=columns 20-27
//bits 3,2,1,0 = conveyor row minus 4 (Add 4 to this value to get actual conveyor row)
//If byte=0 then no conveyor on level

//extradata[1]
//moving platform of 2 bricks
//bits 7,6=left start position ( 00=1,01=5,10=9,11=13)
//bits 5,4=right end position ( 00=17,01=21,10=25,11=29)
//bits 3,2,1,0 = platform row minus 4
//If byte=0 then no platform on level

//extradata[2]..extradata[17]=guardians (start-row0,start-col0,end-row0,end-col0,
//                                        start-row1,start-col1,end-row1,end-col1,...)
//extradata[18]..extradata[21]=room destinations (N,E,S,W)
//extradata[22]..extradata[nn]=variable length list of objects in the room (row,col)
//3 highests bits of row and/or col can be used as object status bits
//object list terminated with $ff

//Room data
byte room1_data[]=		{

				//Map data
				0xff, 0xfc, 0x3f, 0xff, 
				0x80, 0x00, 0x00, 0x01, 
				0x80, 0x00, 0x00, 0x00, 
				0x9f, 0xff, 0xff, 0xe0, 
				0x80, 0x00, 0x00, 0x00, 
				0x80, 0x00, 0x00, 0x00, 
				0x80, 0x00, 0x00, 0x01, 
				0x80, 0x1f, 0xff, 0xff, 
				0x80, 0x00, 0x00, 0x01, 
				0x81, 0x80, 0x00, 0x01, 
				0x80, 0x00, 0x00, 0x00, 
				0x80, 0x00, 0x00, 0x00, 
				0xbe, 0x03, 0xff, 0xf0, 
				0x80, 0x1c, 0x00, 0x00, 
				0x80, 0x00, 0x00, 0x01, 
				0x80, 0x00, 0x00, 0x01, 
				0x80, 0x1f, 0xff, 0xf1, 
				0x00, 0x00, 0x00, 0x01, 
				0x0f, 0xe0, 0x00, 0x01, 
				0x00, 0x00, 0x00, 0x01, 
				0x00, 0x00, 0x00, 0x01, 
				0xff, 0xfc, 0x3f, 0xff, 

				//Extra data
				0x00, 0x00, 0x09, 0x09, 0x09, 0x1d, 0x10, 0x09, 
				0x04, 0x09, 0x04, 0x01, 0x04, 0x1d, 0xff, 0x00, 
				0x00, 0x00, 0x04, 0x02, 0x07, 0x03, 0x01, 0x01, 
				0x02, 0x1c, 0x08, 0x11, 0x0d, 0x05, 0x0f, 0x11, 
				0x11, 0x0b, 0xff, 

				};
				

byte room2_data[]=		{

				//Map data
				0xff, 0xff, 0xff, 0xff, 
				0x80, 0x00, 0x00, 0x01, 
				0x00, 0x00, 0x00, 0x01, 
				0x00, 0x00, 0x00, 0x01, 
				0x01, 0x80, 0x00, 0x01, 
				0x00, 0x00, 0x60, 0x01, 
				0x80, 0x00, 0x00, 0x61, 
				0x98, 0x00, 0x00, 0x01, 
				0x80, 0x0c, 0x00, 0x01, 
				0x80, 0x00, 0x06, 0x01, 
				0x00, 0x00, 0x00, 0x01, 
				0x00, 0x00, 0x00, 0x00, 
				0x00, 0x00, 0x00, 0x00, 
				0x08, 0x00, 0x0e, 0x00, 
				0x80, 0x00, 0x00, 0x00, 
				0x81, 0x80, 0x00, 0x01, 
				0x80, 0x03, 0x00, 0x31, 
				0x80, 0x00, 0x00, 0x01, 
				0x80, 0x00, 0x01, 0x81, 
				0x83, 0x80, 0x00, 0x01, 
				0x80, 0x00, 0x00, 0x01, 
				0xff, 0xff, 0xff, 0xff, 

				//Extra data
				0x00, 0x00, 0x13, 0x01, 0x01, 0x01, 0x0a, 0x1d, 
				0x0a, 0x01, 0x01, 0x13, 0x0b, 0x13, 0x13, 0x09, 
				0x13, 0x1d, 0x00, 0x0a, 0x00, 0x01, 0x02, 0x07, 
				0x02, 0x11, 0x02, 0x1b, 0x0b, 0x0b, 0x0f, 0x14, 
				0x12, 0x1b, 0xff, 

				};


byte room3_data[]=		{

				//Map data
				0xff, 0xff, 0xff, 0xff, 
				0x80, 0x00, 0x00, 0x01, 
				0x80, 0x00, 0x00, 0x01, 
				0x80, 0x00, 0x00, 0x01, 
				0x80, 0x00, 0x00, 0x01,
				0x80, 0x00, 0x00, 0x01, 
				0x80, 0x00, 0x00, 0x01, 
				0x80, 0x00, 0x00, 0x01, 
				0x80, 0x00, 0x00, 0x01, 
				0x80, 0x00, 0x00, 0x01, 
				0x80, 0x00, 0x00, 0x01, 
				0x80, 0x00, 0x00, 0x01, 
				0x80, 0x00, 0x00, 0x01, 
				0x80, 0x00, 0x00, 0x01, 
				0x80, 0x00, 0x00, 0x01, 
				0x80, 0x00, 0x00, 0x01, 
				0x80, 0x00, 0x00, 0x01, 
				0x80, 0x00, 0x00, 0x00, 
				0xff, 0x80, 0x00, 0x00, 
				0x80, 0x00, 0x00, 0x00, 
				0x80, 0x00, 0x00, 0x00, 
				0xff, 0xff, 0xff, 0xff, 

				//Extra data
				0x00, 0x00, 0x01, 0x19, 0x13, 0x19, 0x03, 0x14, 
				0x13, 0x14, 0x05, 0x10, 0x13, 0x10, 0x07, 0x0c, 
				0x13, 0x0c, 0x00, 0x01, 0x00, 0x00, 0x14, 0x01, 
				0xff, 

				};
				

byte room4_data[]=		{

				//Map data
				0x87, 0xff, 0xff, 0xff, 
				0x80, 0x00, 0x00, 0x01, 
				0x80, 0x00, 0x00, 0x01, 
				0xf0, 0x00, 0x00, 0x00, 
				0x80, 0x00, 0xe0, 0x38, 
				0x80, 0x20, 0x00, 0x00, 
				0x80, 0x10, 0x00, 0x00, 
				0x80, 0x08, 0x00, 0x01, 
				0x80, 0x04, 0x00, 0x79, 
				0x00, 0x02, 0x00, 0x01, 
				0x00, 0x01, 0x00, 0x01, 
				0x00, 0x00, 0x83, 0x81, 
				0x00, 0x00, 0x40, 0x01, 
				0xe0, 0x00, 0x20, 0x01, 
				0x80, 0x00, 0x10, 0x01, 
				0x8f, 0x00, 0x08, 0x01, 
				0x80, 0x00, 0x04, 0x01, 
				0x80, 0x00, 0x02, 0x01, 
				0x81, 0xe0, 0x01, 0x01, 
				0x80, 0x00, 0x00, 0x81, 
				0x80, 0x00, 0x00, 0x41, 
				0xff, 0xfc, 0x3f, 0xff, 

				//Extra data
				0x42, 0x00, 0x09, 0x1b, 0x13, 0x1b, 0x0e, 0x02, 
				0x13, 0x02, 0x01, 0x09, 0x01, 0x19, 0xff, 0x00, 
				0x00, 0x00, 0x0c, 0x06, 0x01, 0x05, 0x01, 0x03, 
				0x03, 0x11, 0x05, 0x05, 0x07, 0x14, 0x0d, 0x06, 
				0x10, 0x09, 0x13, 0x01, 0x13, 0x16, 0xff, 

				};


byte room5_data[]=		{

				//Map data
				0xff, 0xff, 0xff, 0xff, 
				0x80, 0x00, 0x00, 0x01, 
				0x80, 0x00, 0x00, 0x01, 
				0x80, 0x00, 0x00, 0x01, 
				0x80, 0x00, 0x00, 0x01, 
				0x80, 0x00, 0x30, 0x31, 
				0x80, 0x1f, 0xc8, 0x31, 
				0x80, 0x0f, 0x88, 0x01, 
				0x80, 0x00, 0x08, 0x01, 
				0x80, 0x00, 0x08, 0x00, 
				0x80, 0x00, 0x3e, 0x00, 
				0x87, 0xff, 0xff, 0xe0, 
				0x80, 0x80, 0x01, 0x00, 
				0x80, 0x80, 0x01, 0x01, 
				0x80, 0x80, 0x01, 0x01, 
				0x80, 0x80, 0x01, 0x01, 
				0x80, 0x80, 0x01, 0x01, 
				0x80, 0x80, 0x01, 0x01, 
				0x80, 0x80, 0x01, 0x01, 
				0x80, 0x80, 0x01, 0x01, 
				0x80, 0x80, 0x01, 0x01, 
				0xff, 0xff, 0xff, 0xff, 


				//Extra data
				0x00, 0x30, 0x04, 0x01, 0x13, 0x01, 0x13, 0x09, 
				0x13, 0x15, 0xff, 0x00, 0x00, 0x00, 0x00, 0x00, 
				0x00, 0x00, 0x00, 0x04, 0x00, 0x00, 0x01, 0x01, 
				0x02, 0x1b, 0x07, 0x1a, 0x09, 0x11, 0x0d, 0x03, 
				0x0e, 0x11, 0x14, 0x1e, 0xff, 

				};
				
						
byte room6_data[]=		{

				//Map data
				0xff, 0xff, 0xff, 0xff, 
				0x80, 0x00, 0x00, 0x01, 
				0x80, 0x00, 0x00, 0x01, 
				0x0f, 0xff, 0x80, 0xe1, 
				0x00, 0x00, 0x00, 0x01, 
				0x00, 0x00, 0x00, 0x01, 
				0x00, 0x01, 0xff, 0xe1, 
				0x90, 0x00, 0x00, 0x01, 
				0x90, 0x00, 0x00, 0x01, 
				0x90, 0x00, 0x00, 0x81, 
				0x90, 0x00, 0x00, 0x01, 
				0x90, 0x00, 0x00, 0x01, 
				0x90, 0x00, 0x00, 0xe1, 
				0x90, 0x00, 0x00, 0x01, 
				0x90, 0x00, 0x00, 0x01, 
				0x9f, 0xff, 0xff, 0xf9, 
				0x9f, 0xff, 0xff, 0xf9, 
				0x90, 0x00, 0x00, 0x09, 
				0x90, 0x00, 0x00, 0x09, 
				0x90, 0x00, 0x00, 0x09, 
				0x90, 0x00, 0x00, 0x09, 
				0xff, 0xff, 0xff, 0xff, 

				//Extra data
				0x00, 0x00, 0x0d, 0x04, 0x0d, 0x1d, 0x13, 0x1d, 
				0x01, 0x1d, 0x04, 0x1d, 0x04, 0x01, 0xff, 0x00, 
				0x00, 0x00, 0x00, 0x00, 0x00, 0x04, 0x01, 0x03, 
				0x01, 0x17, 0x07, 0x0f, 0x0c, 0x07, 0x13, 0x01, 
				0x13, 0x1b, 0xff, 

				};

byte room7_data[]=		{

				//Map data
				0xff, 0xfc, 0x3f, 0xff, 
				0x80, 0x80, 0x02, 0x01, 
				0x80, 0x80, 0x02, 0x01, 
				0x84, 0x90, 0x12, 0x41, 
				0x84, 0x91, 0x92, 0x41, 
				0x83, 0x60, 0x0d, 0x81, 
				0x80, 0x00, 0x00, 0x01, 
				0x80, 0x05, 0x40, 0x01, 
				0x80, 0x05, 0x40, 0x01, 
				0x80, 0x02, 0x80, 0x01, 
				0x80, 0x01, 0x00, 0x01, 
				0x80, 0x01, 0x00, 0x01, 
				0x80, 0x03, 0x80, 0x01, 
				0x8f, 0xff, 0xff, 0xf0, 
				0x82, 0x00, 0x00, 0x40, 
				0x82, 0x00, 0x00, 0x40, 
				0x82, 0x00, 0x00, 0x40, 
				0x02, 0x00, 0x00, 0x41, 
				0x02, 0x00, 0x00, 0x41, 
				0x02, 0x00, 0x00, 0x41, 
				0x05, 0x00, 0x00, 0xa1, 
				0xff, 0xff, 0xff, 0xff, 
				
				//Extra data
				0x00, 0x00, 0x01, 0x01, 0x13, 0x01, 0x0b, 0x01, 
				0x0b, 0x0c, 0x05, 0x12, 0x0b, 0x12, 0xff, 0x00, 
				0x00, 0x00, 0x01, 0x08, 0x00, 0x0d, 0x02, 0x19, 
				0x03, 0x06, 0x0a, 0x06, 0x0c, 0x19, 0x0f, 0x11, 
				0x14, 0x06, 0xff, 

				};


byte room8_data[]=		{

				//Map data
				0xff, 0xff, 0xff, 0xff, 
				0x80, 0x00, 0x00, 0x01, 
				0x80, 0x00, 0x00, 0x01, 
				0x80, 0x00, 0x00, 0x01, 
				0xff, 0xc0, 0x00, 0x01, 
				0x80, 0x00, 0x00, 0x01, 
				0x80, 0x00, 0x00, 0x01, 
				0x8f, 0x00, 0x1f, 0xff, 
				0x89, 0x00, 0x00, 0x01, 
				0x89, 0x00, 0x00, 0x00, 
				0xff, 0xfc, 0x00, 0x00, 
				0x80, 0x00, 0x00, 0x00, 
				0x80, 0xf0, 0x00, 0x00, 
				0x00, 0x0f, 0x00, 0x01, 
				0x00, 0x00, 0x00, 0x01, 
				0x00, 0x01, 0xff, 0x81, 
				0x00, 0x00, 0x00, 0x01, 
				0x80, 0x00, 0x00, 0x01, 
				0x9f, 0xfe, 0x00, 0x41, 
				0x80, 0x00, 0x00, 0x01, 
				0x80, 0x00, 0x00, 0x01, 
				0xff, 0xff, 0xff, 0xff, 

				//Extra data
				0x00, 0x00, 0x02, 0x02, 0x02, 0x1d, 0x08, 0x14, 
				0x0d, 0x14, 0x13, 0x1d, 0x13, 0x01, 0xff, 0x00, 
				0x00, 0x00, 0x00, 0x09, 0x00, 0x07, 0x01, 0x02, 
				0x04, 0x1d, 0x08, 0x05, 0x0d, 0x15, 0x10, 0x06, 
				0x11, 0x19, 0xff, 

				};


byte room9_data[]=		{

				//Map data
				0xff, 0xff, 0xff, 0xff, 
				0x80, 0x00, 0x00, 0x01, 
				0x80, 0x00, 0x00, 0x01, 
				0x80, 0x00, 0x00, 0x01, 
				0x80, 0x00, 0x11, 0x11, 
				0x80, 0x00, 0x09, 0x20, 
				0x81, 0xf0, 0x05, 0x40, 
				0x81, 0x10, 0x03, 0x80, 
				0x81, 0x10, 0x01, 0x00, 
				0x01, 0x10, 0x01, 0x01, 
				0x01, 0xf0, 0x01, 0x01, 
				0x00, 0x00, 0x01, 0x01, 
				0x00, 0x00, 0x01, 0x01, 
				0x82, 0x80, 0x01, 0x01, 
				0x83, 0x80, 0x01, 0x01, 
				0x82, 0x80, 0x01, 0x01, 
				0x82, 0x80, 0x01, 0x01, 
				0x82, 0x80, 0x01, 0x01, 
				0x82, 0x80, 0x01, 0x01, 
				0x82, 0x80, 0x01, 0x01, 
				0x82, 0x80, 0x01, 0x01, 
				0xff, 0xff, 0xff, 0xff, 

				//Extra data
				0x00, 0x00, 0x01, 0x0d, 0x13, 0x0d, 0x0b, 0x06, 
				0x0b, 0x15, 0x01, 0x0f, 0x09, 0x0f, 0x0b, 0x18, 
				0x13, 0x18, 0x00, 0x10, 0x00, 0x08, 0x02, 0x0c, 
				0x03, 0x12, 0x07, 0x08, 0x09, 0x0a, 0x13, 0x1d, 
				0xff, 

				};



byte room10_data[]=		{

				//Map data
				0xff, 0xff, 0xff, 0xff, 
				0x80, 0x00, 0x00, 0x00, 
				0x80, 0x00, 0x00, 0x00, 
				0x80, 0x10, 0x01, 0x00, 
				0x80, 0x10, 0x01, 0x00, 
				0x80, 0x20, 0x00, 0x80, 
				0x80, 0x20, 0x00, 0x80, 
				0x80, 0x40, 0x00, 0x40, 
				0x80, 0x40, 0x00, 0x40, 
				0x8f, 0xfe, 0x0f, 0xfe, 
				0x8a, 0xaa, 0x0a, 0xaa, 
				0x0a, 0xaa, 0x0a, 0xaa, 
				0x0a, 0xaa, 0x0a, 0xaa, 
				0x0a, 0xaa, 0x0a, 0xaa, 
				0x0a, 0xaa, 0x0a, 0xaa, 
				0x8a, 0xaa, 0x0a, 0xaa, 
				0x8a, 0xaa, 0x0a, 0xaa, 
				0x8a, 0xaa, 0x0a, 0xaa, 
				0x8a, 0xaa, 0x0a, 0xaa, 
				0x8f, 0xfe, 0x0f, 0xfe, 
				0x80, 0x00, 0x00, 0x00, 
				0xff, 0xff, 0xff, 0xff, 

				//Extra data
				0x00, 0x00, 0x03, 0x0f, 0x13, 0x0f, 0x03, 0x01, 
				0x03, 0x09, 0x07, 0x0a, 0x07, 0x17, 0xff, 0x00, 
				0x00, 0x00, 0x00, 0x0b, 0x00, 0x02, 0x01, 0x13, 
				0x05, 0x11, 0x06, 0x03, 0x08, 0x1d, 0x0c, 0x09, 
				0x0c, 0x19, 0x0f, 0x17, 0x11, 0x01, 0x11, 0x1d, 
				0x14, 0x08, 0x14, 0x17, 0xff, 

				};


byte room11_data[]=		{

				//Map data
				0xff, 0xff, 0xff, 0xff, 
				0x00, 0x00, 0x00, 0x7d, 
				0x00, 0x00, 0x00, 0x7d, 
				0x00, 0x00, 0x00, 0x7d, 
				0x00, 0x00, 0x00, 0x7d, 
				0x00, 0x00, 0x00, 0x7d, 
				0x00, 0x00, 0x00, 0x7d, 
				0x00, 0x00, 0x00, 0x7d, 
				0x3f, 0xff, 0xff, 0xfd, 
				0x1f, 0xff, 0xff, 0x81, 
				0x0f, 0xff, 0xff, 0x81, 
				0x07, 0xff, 0xff, 0x01, 
				0x07, 0xff, 0xfe, 0x01, 
				0x03, 0xff, 0xf8, 0x01, 
				0x03, 0xff, 0xff, 0xff, 
				0x03, 0xff, 0xff, 0xff, 
				0x03, 0xff, 0xff, 0xff, 
				0x01, 0xff, 0xf0, 0x01, 
				0x01, 0xff, 0xf0, 0x01, 
				0x01, 0xff, 0xf0, 0x01, 
				0x01, 0xff, 0xf0, 0x01, 
				0xff, 0xff, 0xff, 0xff, 

				//Extra data
				0x00, 0x00, 0x0c, 0x17, 0x0c, 0x1d, 0x13, 0x1d, 
				0x13, 0x14, 0x0d, 0x04, 0x13, 0x04, 0x06, 0x06, 
				0x06, 0x17, 0x00, 0x00, 0x00, 0x0a, 0x02, 0x00, 
				0x01, 0x1e, 0x02, 0x0e, 0x03, 0x04, 0x05, 0x14, 
				0x09, 0x02, 0x0b, 0x1a, 0x11, 0x06, 0x12, 0x00, 
				0x12, 0x1b, 0xff, 

				};


byte room12_data[]=		{

				//Map data
				0xff, 0xff, 0xff, 0xff, 
				0x80, 0x20, 0x04, 0x01, 
				0x80, 0x40, 0x02, 0x01, 
				0x80, 0x80, 0x01, 0x01, 
				0x81, 0x00, 0x00, 0x81, 
				0x82, 0x0f, 0xe0, 0x41, 
				0x84, 0x09, 0x20, 0x21, 
				0x88, 0x09, 0x20, 0x11, 
				0x90, 0x09, 0x20, 0x09, 
				0xa0, 0x09, 0x20, 0x05, 
				0xc0, 0x0f, 0xe0, 0x03, 
				0x80, 0x00, 0x00, 0x01, 
				0x80, 0x00, 0x03, 0x81, 
				0x80, 0x00, 0x04, 0x41, 
				0x80, 0x00, 0x08, 0x21, 
				0x80, 0x00, 0x08, 0x21, 
				0x80, 0x03, 0xfe, 0x21, 
				0x80, 0x03, 0xfe, 0x21, 
				0x80, 0x03, 0xfe, 0x21, 
				0x80, 0x03, 0xfe, 0x21, 
				0x80, 0x03, 0xfe, 0x21, 
				0x87, 0xff, 0xff, 0xff, 

				//Extra data
				0x00, 0x00, 0x0e, 0x18, 0x13, 0x18, 0x01, 0x01, 
				0x07, 0x01, 0x07, 0x1d, 0x01, 0x1d, 0xff, 0x00, 
				0x00, 0x00, 0x00, 0x00, 0x04, 0x00, 0x03, 0x05, 
				0x03, 0x1b, 0x05, 0x1b, 0x06, 0x0d, 0x0d, 0x1b, 
				0x0e, 0x0a, 0x12, 0x1d, 0xff, 

				};


byte room13_data[]=		{

				//Map data
				0xff, 0xff, 0xff, 0xff, 
				0x80, 0x00, 0x00, 0x01, 
				0x80, 0x00, 0x00, 0x01, 
				0x80, 0x00, 0x00, 0x01, 
				0x80, 0x80, 0x04, 0x01, 
				0x80, 0x80, 0x04, 0x01, 
				0x80, 0x80, 0x04, 0x01, 
				0x80, 0x80, 0x04, 0x01, 
				0x80, 0x80, 0x04, 0x01, 
				0x80, 0x80, 0x04, 0x01, 
				0x80, 0x80, 0x04, 0x01, 
				0x80, 0x80, 0x04, 0x01, 
				0x80, 0x80, 0x04, 0x01, 
				0x80, 0x80, 0x04, 0x01, 
				0x81, 0xc0, 0x0e, 0x01, 
				0x83, 0xe0, 0x15, 0x01, 
				0x83, 0xe0, 0x24, 0x81, 
				0x03, 0xe0, 0x24, 0x80, 
				0x03, 0xe0, 0x24, 0x80, 
				0x01, 0xc0, 0x24, 0x80, 
				0x00, 0x00, 0x00, 0x00, 
				0xff, 0xff, 0xff, 0xff, 

				//Extra data
				0x00, 0x00, 0x01, 0x0d, 0x13, 0x0d, 0x02, 0x01, 
				0x02, 0x1d, 0x04, 0x16, 0x0c, 0x16, 0x04, 0x06, 
				0x0c, 0x06, 0x00, 0x07, 0x00, 0x0e, 0x01, 0x05, 
				0x01, 0x10, 0x02, 0x18, 0x07, 0x1a, 0x08, 0x0d, 
				0x0f, 0x0f, 0x11, 0x03, 0x12, 0x13, 0x13, 0x1b, 
				0xff, 

				};


byte room14_data[]=		{

				//Map data
				0xff, 0xff, 0xff, 0xff, 
				0x00, 0x02, 0x00, 0x01, 
				0x00, 0x01, 0x81, 0xc1, 
				0x00, 0x00, 0x42, 0x01, 
				0x00, 0x0e, 0x24, 0xe1, 
				0x80, 0x11, 0xa9, 0x11, 
				0xc0, 0x20, 0x52, 0x11, 
				0xa0, 0x40, 0x14, 0x09, 
				0x90, 0x40, 0x7c, 0x01, 
				0x88, 0x00, 0x7c, 0x01, 
				0x84, 0x00, 0x38, 0x01, 
				0x82, 0x00, 0x38, 0x01, 
				0x81, 0x07, 0xff, 0xc1, 
				0x80, 0x80, 0x00, 0x01, 
				0x80, 0x40, 0x00, 0x01, 
				0xb3, 0x20, 0x00, 0x01, 
				0xaa, 0x90, 0x00, 0x01, 
				0xab, 0x08, 0x00, 0x00, 
				0xaa, 0x04, 0x00, 0x00, 
				0xb2, 0x02, 0x00, 0x00, 
				0x80, 0x01, 0x00, 0x00, 
				0xff, 0xff, 0xff, 0xff, 

				//Extra data
				0x00, 0x00, 0x0d, 0x0a, 0x0d, 0x1d, 0x05, 0x0d, 
				0x0a, 0x0d, 0x08, 0x01, 0x0d, 0x01, 0xff, 0x00, 
				0x00, 0x00, 0x00, 0x0d, 0x00, 0x0f, 0x02, 0x0b, 
				0x02, 0x1d, 0x04, 0x16, 0x05, 0x04, 0x0b, 0x1a, 
				0x07, 0x01, 0x0f, 0x11, 0x10, 0x07, 0x11, 0x03, 
				0x11, 0x1a, 0x14, 0x0c, 0xff, 

				};

byte room15_data[]=		{

				//Map data
				0xff, 0xff, 0xff, 0xff, 
				0x80, 0x00, 0x00, 0x00, 
				0x80, 0x3c, 0xc0, 0x00, 
				0x80, 0xef, 0xf0, 0x00, 
				0x83, 0xff, 0xbc, 0x00, 
				0x8f, 0xbf, 0xff, 0x01, 
				0x9f, 0xff, 0xfd, 0xc1, 
				0x9d, 0xfd, 0xff, 0xe1, 
				0x9f, 0xff, 0xe7, 0xa1, 
				0x8e, 0xe7, 0xe0, 0xf1, 
				0x87, 0xc3, 0xf0, 0x21, 
				0x83, 0x81, 0xf8, 0x01, 
				0x80, 0x01, 0xf0, 0x01, 
				0x80, 0x01, 0xf0, 0x01, 
				0x80, 0x01, 0xf0, 0x01, 
				0x80, 0x01, 0xf8, 0x01, 
				0x80, 0x01, 0xf8, 0x01, 
				0x80, 0x03, 0xf8, 0x01, 
				0x80, 0x07, 0xfc, 0x01, 
				0x80, 0x1f, 0xfe, 0x01, 
				0x80, 0x7e, 0x0f, 0x01, 
				0xff, 0xff, 0xff, 0xff, 

				//Extra data
				0xcb, 0x00, 0x10, 0x15, 0x0a, 0x15, 0x0a, 0x18, 
				0x13, 0x18, 0x01, 0x01, 0x11, 0x01, 0xff, 0x00, 
				0x00, 0x00, 0x00, 0x0e, 0x00, 0x00, 0x03, 0x0b, 
				0x04, 0x11, 0x05, 0x09, 0x06, 0x16, 0x07, 0x06, 
				0x07, 0x0e, 0x08, 0x19, 0x09, 0x07, 0x09, 0x16, 
				0x0c, 0x07, 0x14, 0x01, 0x14, 0x0f, 0xff, 

				};

byte room16_data[]=		{

				//Map data
				0xff, 0xff, 0xff, 0xff, 
				0x80, 0x00, 0x00, 0x01, 
				0x80, 0x00, 0x00, 0x01, 
				0x80, 0x00, 0x00, 0x01, 
				0x80, 0x00, 0x00, 0x01, 
				0x00, 0xff, 0xff, 0x01, 
				0x00, 0x80, 0x01, 0x01, 
				0x00, 0xbc, 0x3d, 0x01, 
				0x00, 0xa4, 0x25, 0x01, 
				0x80, 0xa4, 0x25, 0x01, 
				0x80, 0xa4, 0x25, 0x01, 
				0x80, 0xbc, 0x3d, 0x01, 
				0x80, 0x80, 0x01, 0x01, 
				0x80, 0x80, 0x0d, 0x01, 
				0x80, 0x80, 0x01, 0x01, 
				0x80, 0xbc, 0x3d, 0x01, 
				0x80, 0xa4, 0x25, 0x01, 
				0x80, 0xa4, 0x25, 0x01, 
				0x80, 0xa4, 0x25, 0x01, 
				0x80, 0xbc, 0x3d, 0x01, 
				0x80, 0x80, 0x01, 0x01, 
				0xff, 0xff, 0xff, 0xff, 

				//Extra data
				0x00, 0x00, 0x01, 0x06, 0x13, 0x06, 0x13, 0x18, 
				0x01, 0x18, 0x06, 0x0f, 0x13, 0x0f, 0xff, 0x00, 
				0x00, 0x00, 0x00, 0x00, 0x00, 0x09, 0x01, 0x10, 
				0x01, 0x1a, 0x04, 0x0b, 0x04, 0x13, 0x05, 0x02, 
				0x05, 0x1c, 0x06, 0x13, 0x09, 0x0b, 0x0a, 0x04, 
				0x0d, 0x02, 0x0d, 0x0b, 0x0d, 0x10, 0x0f, 0x1b, 
				0x10, 0x0b, 0x11, 0x13, 0x12, 0x09, 0x13, 0x03, 
				0xff, 

				};

//Pointers to room data to allow "random access" to rooms
//End the list with a NULL pointer to facilicate reset_game_objects() function
byte *room_ptr[]={room1_data,room2_data,room3_data,room4_data,
			room5_data,room6_data,room7_data,room8_data,
			room9_data,room10_data,room11_data,room12_data,
			room13_data,room14_data,room15_data,room16_data,NULL};

//Room names (Max 24 characters)
char *room_name[]={	"INSIDE SID'S HOUSE",
			"THE DRAWING ROOM",
			"NELLIE'S SECRET NUT",
			"NUTS LANDING",
			"SID'S QUIET STUDY",
			"MASTER BEDROOM",
			"THE DINING HALL",
			"LIVING ROOM",
			"THE HALLWAY",
			"GUEST QUARTERS",
			"THE BATHROOM",
			"THE ATTIC",
			"THE KITCHEN",
			"THE CONSERVATORY",
			"THE BACK GARDEN",
			"FRONT DOOR"};
	
//Sprites

//16x24 pixel Sprite data follows (sprite data only, no mask data)
//No masks needed for this game, as sprite data is written in "OR" mode
//All sprites and tiles designed and exported using SevenuP (Thanks Metalbrain!)
//Padding of 7 zero bytes manually pre-pended and array size increased from 384
//to 391 bytes to facilitate use with SP1

/* C source file created by SevenuP v1.20                                */
/* SevenuP (C) Copyright 2002-2006 by Jaime Tejedor Gomez, aka Metalbrain*/

/*
GRAPHIC DATA:
Pixel Size:      ( 16,  24)
Char Size:       (  2,   3)
Frames:             8
Sort Priorities: Char line, Y char, X char, Frame number
Data Outputted:  Gfx
Interleave:      Sprite
Mask:            No
*/

//Player sprite (Sid Spanners has 4 right-facing frames and 4 left-facing frames)
//Plus an extra 'blank frame' for invulnerablity
unsigned char sprite_sid1[439] = {

//7 zero data bytes padding
0,0,0,0,0,0,0,

//right0
  0, 14, 31, 29, 31, 14, 63,127,
239,207, 31, 63, 51, 99,113, 48,
  0,  0,  0,  0,  0,  0,  0,  0,
  0,  0, 18,179, 51, 18, 12,156,
252,108, 12, 12, 76,192,192,128,
  0,  0,  0,  0,  0,  0,  0,  0,

//right1
  0, 14, 31, 29, 31, 14, 31, 63,
111,111, 63, 31, 51, 51, 27, 29,
  0,  0,  0,  0,  0,  0,  0,  0,
  0,  0,  0,128,  0,  6, 12,152,
209,219,126,112,224,192,128,192,
  0,  0,  0,  0,  0,  0,  0,  0,

//right2
  0, 14, 31, 29, 31, 14, 15, 15,
 15, 15, 31, 31,  6,  6,  7,  7,
  0,  0,  0,  0,  0,  0,  0,  0,
  0,  0,  0,128,  0,  0,  0,  0,
140,158,224,224, 30, 12,  0,  0,
  0,  0,  0,  0,  0,  0,  0,  0,

//right3
  0, 14, 31, 29, 31,110,127,127,
 63, 31, 15, 31, 27, 51, 61, 29,
  0,  0,  0,  0,  0,  0,  0,  0,
  0,  0,  0,128,  0,  0,  0,  0,
  0,128,224,240,176,144,224,192,
  0,  0,  0,  0,  0,  0,  0,  0,

//left0
  0,  0, 72,205,204, 72, 48, 57,
 63, 54, 48, 48, 50,  3,  3,  1,
  0,  0,  0,  0,  0,  0,  0,  0,
  0,112,248,184,248,112,252,254,
247,243,248,252,204,198,142, 12,
  0,  0,  0,  0,  0,  0,  0,  0,

//left1
  0,  0,  0,  1,  0, 96, 48, 25,
139,219,126, 14,  7,  3,  1,  3,
  0,  0,  0,  0,  0,  0,  0,  0,
  0,112,248,184,248,112,248,252,
246,246,252,248,204,204,216,184,
  0,  0,  0,  0,  0,  0,  0,  0,

//left2
  0,  0,  0,  1,  0,  0,  0,  0,
 49,121,  7,  7,120, 48,  0,  0,
  0,  0,  0,  0,  0,  0,  0,  0,
  0,112,248,184,248,112,240,240,
240,240,248,248, 96, 96,224,224,
  0,  0,  0,  0,  0,  0,  0,  0,

//left3
  0,  0,  0,  1,  0,  0,  0,  0,
  0,  1,  7, 15, 13,  9,  7,  3,
  0,  0,  0,  0,  0,  0,  0,  0,
  0,112,248,184,248,118,254,254,
252,248,240,248,216,204,188,184,
  0,  0,  0,  0,  0,  0,  0,  0,

//blank frame for "invulnerability"
  0,0,0,0,0,0,0,0,
  0,0,0,0,0,0,0,0,
  0,0,0,0,0,0,0,0,
  0,0,0,0,0,0,0,0,
  0,0,0,0,0,0,0,0,
  0,0,0,0,0,0,0,0,
};

//The four guardian sprites, each having 2 frames
unsigned char sprite_guardian1[391] = {
	
//7 zero data bytes padding
	0,0,0,0,0,0,0,

//guardian 0 frame 0
227,227, 65, 79, 95,119,103,  7,
  7, 12, 24, 24, 24, 24, 60, 60,
  0,  0,  0,  0,  0,  0,  0,  0,
192,192,128,240,248,239,231,231,
224, 48, 24, 24, 24, 24, 60, 60,
  0,  0,  0,  0,  0,  0,  0,  0,
  
//guardian 0 frame 1
  3,  3,  1, 15, 31,247,231,231,
  7, 12, 24, 24, 24, 24, 60, 60,
  0,  0,  0,  0,  0,  0,  0,  0,
199,199,130,242,250,238,230,224,
224, 48, 24, 24, 24, 24, 60, 60,
  0,  0,  0,  0,  0,  0,  0,  0,
  
//guardian 1 frame 0  
 31, 63,115,251,255,255,171,  1,
 83,255,255,255,255,127, 63, 31,
  0,  0,  0,  0,  0,  0,  0,  0,
192,240,248,252,252,254,254,254,
254,254,254,252,252,248,240,192,
  0,  0,  0,  0,  0,  0,  0,  0,
  
//guardian 1 frame 1
  3, 15, 31, 63, 63,127,127,127,
127,127,127, 63, 63, 31, 15,  3,
  0,  0,  0,  0,  0,  0,  0,  0,
248,252,206,223,255,255,213,128,
202,255,255,255,255,254,252,248,
  0,  0,  0,  0,  0,  0,  0,  0,
  
//guardian 2 frame 0
  8,  4,  2,255,128,159,191,165,
173,164,182,166,191,159,128,255,
  0,  0,  0,  0,  0,  0,  0,  0,
 16, 32, 64,255,  1,241,251,201,
219, 89,219,201,251,241,  1,255,
  0,  0,  0,  0,  0,  0,  0,  0,
  
//guardian 2 frame 1
  8,  4,  2,255,128,159,167,175,
165,181,164,190,190,159,128,255,
  0,  0,  0,  0,  0,  0,  0,  0,
 16, 32, 64,255,  1,241,203,217,
219,217, 75,249,251,241,  1,255,
  0,  0,  0,  0,  0,  0,  0,  0,
  
//guardian 3 frame 0
 48, 72, 72,112, 48, 24, 13,  7,
  7, 13, 24, 48,112, 72, 72, 48,
  0,  0,  0,  0,  0,  0,  0,  0,
  0, 14, 28, 56,112,224,192,128,
128,192,224,112, 56, 28, 14,  0,
  0,  0,  0,  0,  0,  0,  0,  0,
  
//guardian 3 frame 1
  0,112, 56, 28, 14,  7,  3,  1,
  1,  3,  7, 14, 28, 56,112,  0,
  0,  0,  0,  0,  0,  0,  0,  0,
 12, 18, 18, 14, 12, 24,176,224,
224,176, 24, 12, 14, 18, 18, 12,
  0,  0,  0,  0,  0,  0,  0,  0
};

//8x8 tiles (bricks, nuts and door)
byte tile_brick0[] = {223,223,223,  0,251,251,251,  0};
byte tile_nut[] = { 0, 60,102, 66, 66,102, 60,  0};


//Nellie nutjob
byte tile_nellie_k[] = {12, 31, 55, 55,103,  3, 31, 63};
byte tile_nellie_l[] = {119,103,111, 31, 63,  6, 14, 28};
byte tile_nellie_m[] = {0,128,192, 96,192,135,243,255};
byte tile_nellie_n[] = {238,224,240,248,252, 96,112, 56};

byte tile_nellie_o[] = {12, 31, 55,119, 71,  3, 31, 63};
byte tile_nellie_p[] = {55, 63, 31, 63,127,  6, 30, 56};
byte tile_nellie_q[] = {0,128,192, 96,192,128,240,254};
byte tile_nellie_r[] = {239,227,240,240,248, 96, 96,112};

//conveyor belt (8 x 8x8 rotated tiles)
byte conveyor_tile[64];

//Function prototypes
void *u_malloc(uint size);
void u_free(void *addr);
void init_universe(void);
void init_conveyor_tiles(void);
void init_im2_handler(void);
void init_keyboard(void);
void intro_screen(void);
void init_game(void);
void init_room(void);
void init_movers(void);
void define_tileset(void);
void draw_room_background(void);
void init_sprites(void);
void zx_border3(byte b);
void play_tone(word cycles, word length, word delta);
void play_white_noise(word cycles, word length, word delta);
//Custom print and cls routines (not using stdio.h in this game)
//If attr is zero, then write no attributes (use existing)
void my_print_at(byte row, byte col,char *string, byte attr);
void my_cls(byte attr);
void my_borcls(byte attr,byte b);
void init_anim_lists(void);
void wtod(word value, char* str, byte len);
void reverse_string(char* begin, char* end);
void add_sprite_colour(struct sp1_cs *cs);
void calculate_sprite_colours(void);
void draw_status_area(void);
void draw_room_name(void);
void draw_lives(void);
void draw_nuts(void);
void reset_game_objects(void);
void game_over(void);
void won_game(void);
void collect_object(void);
void lose_a_life(void);
void play_mrbeep_tune(void);

main()
{
//loop variables
byte i;
char j;

//immediately disable interrupts, to prevent IM1 routine from clobbering sysvar area
#asm
di
#endasm

//Run this once and only once
init_universe();

beginning:

while(1)
	{
	intro_screen();

	init_game();

	//debugging
	//nuts=1;
	//lives=1;

enter_room:

	//room_exited can be set to 3 to signal to init_room() that a new room has been entered

	init_room();

	//reset room_exited variable on entering room	
	room_exited=0;

	while(!room_exited)
		{

		//Main game loop


		//Update screen
		sp1_UpdateNow(); 
		
		//Reset the flag which checks if
		//the player has been caught by an enemy.
		caught=0;

		//Calculate absolute position of all sprites
		for(i=0;i<guardians;i++)
			{
			se=&sprtbl[i];
			se->px=(se->s->col<<3)+(se->s->hrot&7);
			se->py=(se->s->row<<3)+(se->s->vrot&7);
			
			//collision detection
			//for each enemy sprite - detect if the player
			//sprite has been touched
			if(i)
				{
				//This old code had a bug
				//need (int) casts to produce signed result
				//cx=(byte)ABS((int)(se->px)-(int)(sprtbl[0].px));
				//cy=(byte)ABS((int)(se->py)-(int)(sprtbl[0].py));

				//Fixed code in v1.1- doesn't need any casts
				//This does same as above was intended to, except that it works correctly 
				//This fix backported from Sid Spanners 4
				cx=(se->px>sprtbl[0].px?se->px-sprtbl[0].px:sprtbl[0].px-se->px);
				cy=(se->py>sprtbl[0].py?se->py-sprtbl[0].py:sprtbl[0].py-se->py);

				//These values dictate the sensitivity of the collision detection
				if((cx<12)&&(cy<12))
					caught=i;
				}			

			}

		//Handle Player Character
		//Set sprite entry pointer to sprite zero (player sprite)
		se = &sprtbl[0];

		//Check for exiting the room
		if((se->px==0)||(se->px==240))
			{
			//exited horizontally - set travel direction
			travel_dir=(se->px==240?'e':'w');
			//extract either "east" or "west" destination from "extra data"
			dest_room=room_data[X_OFFSET+(travel_dir=='e'?19:21)];
			room_exited=3; //3 is the code for exiting the edge of a room
			break; //break out of while loop (no need to process this frame if exiting room)
				//in fact this break statement is needed to allow jumping through doors
				//without the "wraparound" wall collision detection resssetting se->dx to 0
			}

		if((se->py<2)||(se->py>173))
			{
			//exited vertically
			//North or South?
			travel_dir=(se->py<2?'n':'s');
			dest_room=room_data[X_OFFSET+(travel_dir=='n'?18:20)];
			room_exited=3;
			break;
			}

		//check for touching an object
		//Reset 'touched' variable to no item
		touched=' ';

		//disregard checking when player has jumped off the top of the map
		if(se->py<192)
			{
			//check six squares behind the player (1 row and 2 rows above player feet)
			for(i=0;i<3;i++)
				for(j=0;j<2;j++)
					{
					my_attr=MY_ATTR(se->s->row+j,se->s->col+i);

					//Check for nuts
					if(my_attr==(INK_WHITE|BRIGHT))
						{
						ty=se->s->row+j;
						tx=se->s->col+i;
						//'b' is internal symbol for nut
						touched='b';		
						}			
					}
			}

		//deal with touched nut (enough of that - ed)
		if(touched=='b')
			{
			sp1_PrintAt(ty,tx,PAPER_BLACK|INK_YELLOW|BRIGHT,' ');
			onecr.row=ty;
			onecr.col=tx;
			sp1_Invalidate(&onecr);
			sp1_UpdateNow();
			nuts--;
			draw_nuts();
			play_tone(100,200,-2);
			collect_object();

			//Game won if no nuts remaining
			if(!nuts)
				{
				//game won
				room_exited=2;
				}
			}

		if(invulnerable)
			{
			invulnerable--;
			}
		else
			{
			//deal with player caught, but only if not invulnerable
			if(caught)
				{
				lose_a_life();
				//deal with player dead
				if(!lives)
					{
					//set room_exited flag to 1 when player dies
					room_exited=1;
					}			
				}
			}

			//Process generic room specific mods
			switch(room)
				{
				case 3:	//Nellie's secret nut
					//only show nellie when there is more than one nut remaining
					if(nuts>1)
						{
						if(guardian_flip)
							{
							sp1_PrintAt(19,7,0x42,'k');
							sp1_PrintAt(20,7,0x42,'l');
							sp1_PrintAt(19,8,0x42,'m');
							sp1_PrintAt(20,8,0x42,'n');
							}
						else
							{
							sp1_PrintAt(19,7,0x42,'o');
							sp1_PrintAt(20,7,0x42,'p');
							sp1_PrintAt(19,8,0x42,'q');
							sp1_PrintAt(20,8,0x42,'r');
							}
						sp1_Invalidate(&nellie_cr);	
						//Check for player in "Illegal area" when Nellie present
						if((se->px<70)&&(se->py>144))
							lose_a_life();
							
							//v1,2 Fix the "Nellie Nutjob death bug"

							//deal with player dead
							if(!lives)
							{
							//set room_exited flag to 1 when player dies
							room_exited=1;
							}	
						}
				break;
				//for most rooms, do nothing
				default:
				break;
				}

		//detect a solid surface under player
        	//only check the last tile if the horizontal rotation is non-zero
		//"and" the attribute with 0x47 as we are only checking the INK colour (and BRIGHT)
            	solid_ground=((MY_ATTR(se->s->row+2,se->s->col)&0x47)==brick_colour)||
			((MY_ATTR(se->s->row+2,se->s->col+1)&0x47)==brick_colour)||
			(((MY_ATTR(se->s->row+2,se->s->col+2)&0x47)==brick_colour)&&se->s->hrot);

		//check for walking off the end of a platform
		if((!solid_ground)&&(!airborne))
			{
			se->dy=1;
			airborne=1;
			}

		//read keyboard 
		input = (getjoy)(&keys); 

        	//player walking on a platform
		if(!airborne)
			{
			//check when walking left that there is no "solid wall" (with mortar) one pixel to the left of the 
			//player. Check PAPER colour
			if((input & in_LEFT)&&((MY_ATTR(se->s->row+1,(se->px-1)>>3)&0x38)!=mortar_colour)) 
				{
				facing_left=1;				
				se->dx=-1;
				
				//if somehow got stuck inside a platform, then
				//move back onto the top of it
				if(se->s->vrot&7)
					{
					se->dy=-(se->s->vrot&7);
					}

				//animate sprite

				if(frame_tick++>1)	
					{
					se->frame++;
					if(se->frame>5) se->frame=0;
					frame_tick=0;
					}
				}

			if((input & in_RIGHT)&&((MY_ATTR(se->s->row+1,(se->px+16)>>3)&0x38)!=mortar_colour))
				{
				facing_left=0;
				se->dx=1;

				//if somehow got stuck inside a platform, then
				//move back onto the top of it
				if(se->s->vrot&7)
					{
					se->dy=-(se->s->vrot&7);
					}

				//animate sprite
				if(frame_tick++>1)	
					{
					se->frame++;
					if(se->frame>5) se->frame=0;
					frame_tick=0;
					}
				}
			if(input & (in_FIRE|in_UP|in_DOWN))
				{
				airborne=1;
				se->dx=(input & in_RIGHT?1:0)-(input & in_LEFT?1:0);
				se->dy=-2;
				}			
			}

		if(airborne)
			{
			//adjust speed every few ticks
            		//The higher this value, the higher sid can jump
			if(jump_tick++>8)
				{
				if(se->dy<2) se->dy++;
				jump_tick=0;
				//if speed changed to +2 on an odd line (during falling), then adjust to +1
				//for one more iteration, to keep the vertical position an
				//even number and avoid landing inside a platform
				if((se->dy==2)&&(se->s->vrot&1))
					{
					se->dy=1;
					//force increment to speed of two on following interation
					jump_tick=9;
					}
				}

			//landed - check for sold ground, moving downwards and player not off the top of the screen
			//also check for vrot values of 0,1,2 preventing player getting stuck inside platform
			//near the top of the jump
			i=se->s->vrot&7;
            		if(solid_ground&&(se->dy>=0)&&(se->py<192)&&(i<3)) 
				{
				airborne=0;
				//nullify residual speed to land exactly on platform				
				se->dy=0;
				jump_tick=0;
				//if airborne_entry is set, then reset "lost life return co-ordinates"
				// to the first platform landed on
				if(airborne_entry)
					{
					se->startrow=se->s->row;
					se->startcol=se->s->col;

					se->endcol=se->s->hrot;
					se->endrow=se->s->vrot;

					airborne_entry=0;
					}
				}
            		//prevent jumping horizontally through solid walls
			if((se->dx==-1)&&((MY_ATTR(se->s->row+1,(se->px-1)>>3)&0x38)==mortar_colour)) se->dx=0;
			if((se->dx==1)&&((MY_ATTR(se->s->row+1,(se->px+16)>>3)&0x38)==mortar_colour)) se->dx=0;

			//prevent jumping through the ceiling
			if((se->dy<0)&&(((MY_ATTR((se->py-1)>>3,se->s->col)&0x38)==mortar_colour)||
					(MY_ATTR((se->py-1)>>3,se->s->col+1)&0x38)==mortar_colour)) se->dy=0;

			//Hand-coded adjustments for some non-standard rooms
			//(Affecting airborne player)
			switch(room)
				{				
				//for most rooms, do nothing
				default:
				break;
				}
			}  //end if(airborne)

		//Handle conveyor belt
		if(conveyor)
			{
			//print sections of conveyor belt which are enabled
			for(i=0;i<8;i++)
				{
				if(conveyor&0x40) sp1_PrintAt(convrow,4+i,brick_colour,'c'+convpos);
				if(conveyor&0x20) sp1_PrintAt(convrow,12+i,brick_colour,'c'+convpos);
				if(conveyor&0x10) sp1_PrintAt(convrow,20+i,brick_colour,'c'+convpos);
				}
			//invalidate conveyor area
			sp1_Invalidate(&conveyor_cr);
			//move conveyor left or right
			if(conveyor&0x80)
				convpos--;
			else
				convpos++;

			convpos&=7;

			//handle player standing on conveyor
			if((!airborne)&&(se->s->row==(convrow-2)))
				{
				//Version 1.1 This additional "if" statement fixes the "conveyor tree bug"
				//Bugfix backported from Sid Spanners 4
				if(!((room_data[(convrow<<2)+(se->s->col>>3)]&bit_selector[se->s->col&0x07])||
				(room_data[(convrow<<2)+((se->s->col+1)>>3)]&bit_selector[(se->s->col+1)&0x07])||
				((room_data[(convrow<<2)+((se->s->col+2)>>3)]&bit_selector[(se->s->col+2)&0x07])&&se->s->hrot)))
					{
					if(conveyor&0x80)
						{
						//conveyor moving left
						if(se->dx>-1) se->dx--;
						}
					else
						{
						//conveyor moving right
						if(se->dx<1) se->dx++;
						}
					}
				}
			}

		//Handle moving platform
		if(moveplat)
			{
			//Move platform when timer elapses
			if(!mptick++)
				{
				sp1_PrintAt(mprow,mpcol-1,INK_YELLOW|PAPER_BLACK|BRIGHT,' ');
				sp1_PrintAt(mprow,mpcol,brick_colour,'a');
				sp1_PrintAt(mprow,mpcol+1,brick_colour,'a');
				sp1_PrintAt(mprow,mpcol+2,INK_YELLOW|PAPER_BLACK|BRIGHT,' ');

				if(mpleft)
					{
					//moving left
					mpcol--;
					if(mpcol==mpstartcol)
						mpleft=0;
					}
				else
					{
					//moving right
					mpcol++;
					if(mpcol==mpendcol)
						mpleft=1;
					}
				//Invalidate moving platform area
					sp1_Invalidate(&moveplat_cr);
				}
			else
				{
				if(mptick>MOVEPLAT_DELAY) mptick=0;
				}
			}
		 
		// move/draw player sprite
		//if invulnerable, then flicker sprite
		if(invulnerable&2)
			sp1_MoveSprRel(se->s, &cr, sprite_sid1+391, 0, 0, se->dy, se->dx); 		
		else
			{
			if(facing_left)
				sp1_MoveSprRel(se->s, &cr, anim_sid_left[se->frame], 0, 0, se->dy, se->dx); //Move sprite
			else
				sp1_MoveSprRel(se->s, &cr, anim_sid_right[se->frame], 0, 0, se->dy, se->dx); 
         		}

		if(!airborne)
			{
			//stop moving if not airborne
			se->dx=0;
         		se->dy=0;
			}

		//move guardian sprites
		for(i=1;i<guardians;i++)
			{
			//set up se to point to current sprite entry
			se=&sprtbl[i];
			if(se->startrow==se->endrow)
				{
				//horizontal guardian
				if(se->direction==1)
					{
					
					//going from start to end
					//cast operands to signed values
					se->dx=SGN((char)se->endcol-(char)se->startcol);

					//check for end reached
					if((se->s->col==se->endcol)&&(se->s->hrot==0))
						{
						se->direction=-1; //reverse direction
						se->dx=0; //stop moving
						}
					}
				else
					{
					//going from end to start
					//cast operands to signed values
					se->dx=SGN((char)se->startcol-(char)se->endcol);
					//check for start reached
					if((se->s->col==se->startcol)&&(se->s->hrot==0))
						{
						se->direction=1; //forward direction
						se->dx=0; //stop moving
						}
					}
				}
			else
				{
				//vertical guardian
				if(se->direction==1)
					{
					//going from start to end
					//cast operands to signed values
					se->dy=SGN((char)se->endrow-(char)se->startrow);

					//check for end reached
					if((se->s->row==se->endrow)&&((se->s->vrot&7)==0))
						{
						se->direction=-1; //reverse direction
						se->dy=0; //stop moving
						}
					}
				else
					{
					//going from end to start
					//cast operands to signed values
					se->dy=SGN((char)se->startrow-(char)se->endrow);
					//check for start reached
					if((se->s->row==se->startrow)&&((se->s->vrot&7)==0))
						{
						se->direction=1; //forward direction
						se->dy=0; //stop moving
						}
					}
											
				}  //end else

				//Room-specific guardian behaviours mods
				switch(room)
					{
					//for most rooms, do nothing
					default:
					break;
					}

			//move the sprite	
			sp1_MoveSprRel(se->s, &cr, (guardian_flip?start_frame[i]+48:start_frame[i]), 0, 0, se->dy, se->dx); 		
			} //end for()


			//alternate between two animated frames
			if(guardian_tick++>15)
				{			
				guardian_flip^=1;
				guardian_tick=0;
				}
		} //end while(!room_exited)

	//This point is reached if the game is over or if the room has been exited

	//Exited conditions
	//room_exited==1 Player has ran out of lives (game over)
	//room_exited==2 Player has collected all nuts (won the game)
	//room_exited==3 Player has exited via an exit (enter the desination room)

	if(room_exited==1)
		{
		//game over
		for(i=0;i<6;i++)
			play_tone(50,100,8);
		game_over();		
		}
	else if(room_exited==2)
		{
		//won game
		for(i=180;i<200;i++)
			play_tone(i,200-i,0);
		won_game();		
		}
	else if(room_exited==3)
		{
		//set new room to destination
		room=dest_room;
		goto enter_room;
		}


	} //end while(1)
} //end main()

void *u_malloc(uint size) 
{
return malloc(size);
}

void u_free(void *addr)
{
free(addr);
}


void init_universe(void)
{
struct sp1_ss *s; 
byte i;

//Run this immediately - Replace system ISR routine and enter IM2 mode
   init_im2_handler();

//Last minute game hacks go here
//stop the "O" in the font looking like a "D"
bpoke(53117,0x62);

//v1.2 fix the "Intro screen auto-exit bug"
//Patch the player code such that it doesn't exit immediately on first run 
//on the WoS Java emulator ZZSpectrum

//Remove Kempston detection code
bpoke(0xc00f,0xaf);
bpoke(0xc010,0xaf);
bpoke(0xc044,0xaf);
bpoke(0xc045,0xaf);


// Setup the heap
   mallinit();              // set heap as currently empty
   sbrk(HEAP_ADDR,HEAP_SIZE);        // Allocate defined chunk to heap

//Initialise SP1 sprite engine
	sp1_Initialize(SP1_IFLAG_MAKE_ROTTBL | SP1_IFLAG_OVERWRITE_TILES | SP1_IFLAG_OVERWRITE_DFILE,
	 INK_YELLOW|PAPER_BLACK|BRIGHT, ' ');


//Create the sprites. In this game, they are maskless 1-byte sprites of the "OR" draw type
	for(i=0;i<NUM_SPRITES;i++)
		{
		s=sprtbl[i].s = sp1_CreateSpr(SP1_DRAW_OR1LB, SP1_TYPE_1BYTE, 3, 0, i);

      		sp1_AddColSpr(s, SP1_DRAW_OR1, 0, 24, i);
      		sp1_AddColSpr(s, SP1_DRAW_OR1RB, 0, 0, i);
		}

//Initialise the anim lists
init_anim_lists();

//Initialise the rotated tileset used by the conveyor belt
init_conveyor_tiles();

//Initialise Nellie Nutjob tiles
sp1_TileEntry('k',tile_nellie_k);
sp1_TileEntry('l',tile_nellie_l);
sp1_TileEntry('m',tile_nellie_m);
sp1_TileEntry('n',tile_nellie_n);
sp1_TileEntry('o',tile_nellie_o);
sp1_TileEntry('p',tile_nellie_p);
sp1_TileEntry('q',tile_nellie_q);
sp1_TileEntry('r',tile_nellie_r);

}

void init_conveyor_tiles(void)
{
byte tilenum,tilerow,temp;

//Use "brick 0" for the conveyor belt
memcpy(conveyor_tile,tile_brick0,8);

//generate tiles 1-7 by successively rotating tile 0
for(tilenum=1;tilenum<8;tilenum++)
	for(tilerow=0;tilerow<8;tilerow++)
		{
		temp=conveyor_tile[((tilenum-1)<<3)+tilerow];
		temp=(temp>>1)|(temp&1?0x80:0);
		conveyor_tile[(tilenum<<3)+tilerow]=temp;
		}
}

void init_im2_handler(void)
{
#asm  //Disable interrupts
di
#endasm

//Use ISR jump table in standard locations reserved by SP1 library
im2_Init(0xd000);
memset(0xd000,0xd1,257); //Fill 257-byte table with 0xd1d1 

//In this game, the ISR does nothing, so just return immediately 
//(Do not push/pop registers) - poke "RETI" instruction at start
//of ISR routine
bpoke(0xd1d1,0xed);
bpoke(0xd1d2,0x4d);

#asm   //Re-enable interrupts
ei
#endasm

}

void init_keyboard(void)
{
//assign keyboard input keys
getjoy = in_JoyKeyboard; //Use keyboard
keys.left = in_LookupKey('o'); //Left
keys.right = in_LookupKey('p'); //Right
keys.up = in_LookupKey('m'); // Jump
keys.down = in_LookupKey('q'); //Also jump 
keys.fire = in_LookupKey('z'); // Also jump
}

void intro_screen(void)
{
//debugging
//word t,l;
//mallinfo(&t,&l);

//cls all black initially
my_borcls(PAPER_BLACK|INK_BLACK,BLACK);

//cls bright white on black
my_borcls(PAPER_BLACK|INK_WHITE|BRIGHT,BLACK);

my_print_at(2,2,"SID SPANNERS 3: THE NUTHOUSE",0);
my_print_at(4,5,"BY DIGITAL PRAWN 2011",0);

my_print_at(6,3,"INTRO MUSIC BY MISTER BEEP",0);

my_print_at(8,10,"VERSION 1.2",BRIGHT|PAPER_YELLOW|INK_BLUE|FLASH);

//This line was used in the pre-release verion only
//my_print_at(8,6,"PRE-RELEASE VERSION",INK_YELLOW|PAPER_RED|BRIGHT|FLASH);

my_print_at(11,5,"O      ...      LEFT",BRIGHT|INK_GREEN);
my_print_at(13,5,"P      ...      RIGHT",BRIGHT|INK_GREEN);
my_print_at(15,5,"Q/Z/M  ...      JUMP",BRIGHT|INK_GREEN);

//debugging
//wtod(123,tempstr,4);
//my_print_at(6,4,tempstr,0);
//wtod(456,tempstr,6);
//my_print_at(8,4,tempstr,0);

//debugging
//sprintf(tempstr,"T=%d L=%d",t,l);
//my_print_at(18,4,tempstr,0);

//Wait a short while before playing the tune (to prevent unwanted key detection)
in_Wait(750);

my_print_at(20,6,"PRESS A KEY TO START",PAPER_BLACK|INK_WHITE|BRIGHT|FLASH);

play_mrbeep_tune();
//WAIT_KEY;
}

void init_game(void)
{
init_keyboard();

//reset in-game variables
room=START_ROOM;
lives=NUM_LIVES;

//reset in-game objects
reset_game_objects();

//reset player sprite parameters
airborne=0;
facing_left=0;
frame_tick=0;
jump_tick=0;
airborne_entry=0;

//Signal to init_room() that a new game has started
room_exited=0;
}

void init_room(void)
{
//set global pointer to room data for current room
room_data=room_ptr[room-1];

define_tileset();

draw_room_background();

calculate_sprite_colours();

init_sprites();

//Draw the status area
draw_status_area();

//init conveyor belt and moving platform (if present)
init_movers();

//reset guardian animation & action timers
guardian_flip=0;
guardian_tick=0;

//set airborne_entry flag if entering the room whilst airborne
airborne_entry=airborne;

//reset invulnerable flag
invulnerable=0;

}

void init_movers(void)
{

//Check for conveyor belt and set variables accordingly
if(conveyor=room_data[X_OFFSET])
	{
	//calculate conveyor row
	convrow=(conveyor&0x0f)+4;
	//Set clipping rectangle for conveyor belt
	conveyor_cr.row=convrow;
	//reset conveyor belt position
	convpos=0;	
	}

//Check for moving platform and set variables accordingly
if(moveplat=room_data[X_OFFSET+1])
	{
	//calculate moving platform row
	mprow=(moveplat&0x0f)+4;
	//calculate start and end columns
	mpcol=mpstartcol=(((moveplat&0xc0)>>6)<<2)+1;
	mpendcol=(((moveplat&0x30)>>4)<<2)+17;
	//always start by moving right	
	mpleft=0;
	//reset platform move timer
	mptick=0;
	//Set clipping rectangle fort moving platform
	moveplat_cr.row=mprow;
	}
}

void define_tileset(void)
{
byte i;
sp1_TileEntry('a', tile_brick0);
sp1_TileEntry('b', tile_nut);

//Conveyor belt uses 8 tile definitions, beginning at 'c'
for(i=0;i<8;i++)
	sp1_TileEntry('c'+i,&conveyor_tile[i<<3]);
}

void draw_room_background(void)
{
byte row,col;

//Pointer to object data within room data
byte *obj_data;

//Set brick and mortar colours depending on room number
brick_colour=1+(room&3);

mortar_colour=brick_colour-1;
if(mortar_colour<1)
	mortar_colour=4;
	
//Make brick colour BRIGHT
brick_colour|=BRIGHT;

//Shift mortar_colour left by 3 (it is a PAPER colour)
mortar_colour<<=3;

//Redraw all tiles in the play area
for(row=0;row<22;row++)
	for(col=0;col<32;col++)
		{
		//Check for "wall bit"
		if(room_data[(row<<2)+(col>>3)]&bit_selector[col&7])
				{
				//Walls around the edges are "impassable"
				//Use mortar_colour PAPER to show this
				if((row==0)||(row==21)||(col==0)||(col==31))
					{
					sp1_PrintAt(row,col,brick_colour|mortar_colour,'a');
					}
				else
					{
					sp1_PrintAt(row,col,brick_colour,'a');
					}
				}
			else
				sp1_PrintAt(row,col,INK_YELLOW|BRIGHT,' ');
		}

//Now print all objects
	//set pointer to object data
	obj_data=room_data+O_OFFSET;
	
	//continue until "end of data" marker found
	while(*obj_data!=0xff)
		{
		//Currently only one object type supported
		// a "nut"
		row=*obj_data++;
		col=*obj_data++;

		//only print the object if it hasn't been collected
		if((row&0x80)==0)
			sp1_PrintAt(row,col,INK_WHITE|BRIGHT,'b');
		}

//Invalidate entire map
sp1_Invalidate(&cr);

//Temporary update of screen for debugging only - this should be removed
//in normal version as it causes temporary redrawing of sprites in wrong
//positions (from previous room)
//sp1_UpdateNow();

}

void init_sprites(void)
{
//loop variable
byte i;
//initial rotation values of sprites
byte hr,vr;

//initially assume all guardians present
guardians=NUM_SPRITES;

//reset sprite parameters
	for(i=0;i<NUM_SPRITES;i++)
		{

		//Place sprites in their initial positions
		if(i)
			{
			//Guardian sprites
			sprtbl[i].dx=0;
			sprtbl[i].dy=0;
			sprtbl[i].frame=0;
			sprtbl[i].direction=1;
			
			sprtbl[i].startrow=room_data[X_OFFSET+(i<<2)-2];
			sprtbl[i].startcol=room_data[X_OFFSET+(i<<2)-1];
			sprtbl[i].endrow=room_data[X_OFFSET+(i<<2)];
			sprtbl[i].endcol=room_data[X_OFFSET+(i<<2)+1];
			//Colour the guardian sprites
			global_n=i;
			sp1_IterateSprChar(sprtbl[i].s,add_sprite_colour);
			//zero rotations for guardian initial positions
			hr=0;
			vr=0;

			//Check if this is a real guardian or an "end of guardians" marker
			if(sprtbl[i].startrow==0xff)
				guardians=i;
			}
		else
			{
			//Player sprite
			if(room_exited)
				{
				//entered the room via a hole
				switch(travel_dir)
					{
					//travelling north (upwards)
					case 'n':
						sprtbl[i].startrow=19;
						sprtbl[i].startcol=sprtbl[i].s->col;
						hr=sprtbl[i].s->hrot;
						vr=7;
						break;
					//travelling east
					//place sid one pixel from screen edge to prevent
					//returning back to previous screen again 
					case 'e':
						sprtbl[i].startrow=sprtbl[i].s->row;
						sprtbl[i].startcol=0;
						hr=1;				
						vr=sprtbl[i].s->vrot;		
						break;
					//travelling south (downwards)
					case 's':
						sprtbl[i].startrow=0;
						sprtbl[i].startcol=sprtbl[i].s->col;
						hr=sprtbl[i].s->hrot;
						vr=2;
						break;
					//travelling west
					case 'w':
						sprtbl[i].startrow=sprtbl[i].s->row;
						sprtbl[i].startcol=29;
						hr=7;
						vr=sprtbl[i].s->vrot;
						break;
					default:
					break;
					}
				}
			else
				{
				//must be start of game - fix player position at preditermined
				//starting point
				//Defaults (room 1)=(row 19,col 29)
				sprtbl[i].startrow=STARTING_ROW;
				sprtbl[i].startcol=STARTING_COL;
				hr=0;vr=0;
				}
			//A nasty trick - we need somewhere to store the initial hrot and vrot values
			//so that when the player loses a life they can be put back to exactly the same position where 
			//they entered the room. Since the endrow and endcol fields of the sprite structure are
			//unused in the zeroth (player) sprite only, I use them to store the initial rotation values
			sprtbl[0].endcol=hr;
			sprtbl[0].endrow=vr;
			}
		//Place the sprite on-screen (if it is present) else hide it off-screen
		if(i<guardians)
			sp1_MoveSprAbs(sprtbl[i].s, &cr, start_frame[i],sprtbl[i].startrow , sprtbl[i].startcol, vr, hr);
		else
			sp1_MoveSprAbs(sprtbl[i].s, &cr, start_frame[i],30 , 30, 0, 0);

		//debugging
		//wtod(sprtbl[i].startrow,tempstr,3);
		//wtod(sprtbl[i].startcol,tempstr+16,3);
		//my_print_at(i,0,tempstr,0);
		//my_print_at(i,4,tempstr+16,0);
		}

//Temp placeholder function call to allow program to compile
//sp1_UpdateNow();
}

void zx_border3(byte b)
{
zx_border(b);
*(byte *)MY_BORDCR=b;
}

void play_tone(word cycles, word length, word delta)
{
#asm
	di	;disable interrupts
	pop bc  ;pop the return address off the stack
	exx 	;use alternate register set

	pop bc 	;bc holds 'delta'
	pop hl 	;hl holds 'length'
	pop de 	;de holds number of cycles to play


.PlaySFX

;Play Sound Effects
;de holds number of cycles to play
;hl holds the length of each cycle
;bc holds the amount to change the length by on each iteration
.PlaySFXLoop1
    ld a,(MY_BORDCR) ;load current border colour
    xor $10    		;flip the speaker
    out ($fe),a

    push hl; hold amplitude for hl cycles
.PlaySFXLoop2
    dec hl
    ld a,h
    or l
    jr nz,PlaySFXLoop2
    pop hl
     
    ld a,(MY_BORDCR);flip the speaker back
    out ($fe),a
    
    push hl; hold amplitude for hl cycles
.PlaySFXLoop3
    dec hl
    ld a,h
    or l
    jr nz,PlaySFXLoop3
    pop hl

    add hl,bc ;vary the cycle length if bc is non-zero

    dec de
    ld a,d
    or e
    jr nz,PlaySFXLoop1  ; do it 'de' times

    push de  ; restore (changed) parameters to stack
    push hl
    push bc

    exx ; back to original register set
    push bc ; restore return address

    ei

#endasm
}

void play_white_noise(word cycles, word length, word delta)
{
#asm
	di	;disable interrupts
	
	ld hl,$d0d0			;set random seed
	ld (WhiteNoiseGenStart+1),hl

	pop bc  ;pop the return address off the stack
	exx 	;use alternate register set

	pop bc 	;bc holds 'delta'
	pop hl 	;hl holds 'length'
	pop de 	;de holds number of cycles to play

;Play White Noise
;de holds number of cycles to play
;hl holds the length of each cycle
;bc holds the amount to change the length by on each iteration
.PlayWNLoop1
    call WhiteNoiseGenerator ;get random byte in A
    and $10 ;mask off all bits except the speaker bit

    push hl  
    ld hl,MY_BORDCR
    or (hl) ; "or" the value with the current border colour
    pop hl

    out ($fe),a ;send the random bit (with correct border bits)

    push hl; hold amplitude for hl cycles
.PlayWNLoop2
    dec hl
    ld a,h
    or l
    jr nz,PlayWNLoop2
    pop hl

    add hl,bc ;vary the cycle length if bc is non-zero

    dec de
    ld a,d
    or e
    jr nz,PlayWNLoop1  ; do it 'de' times

    push de  ; restore (changed) parameters to stack
    push hl
    push bc

    exx ; back to original register set
    push bc ; restore return address

    ld a,(MY_BORDCR)
    out ($fe),a ;reset speaker bit to 0 before exiting 
                ; (prevent click on next border change)

    jp PlayWNEnd ; jump to end of function

;16-bit pseudo-random generator subroutine
;Modified just to return a random 8-bit value in A
;This random generator subroutine, credited to Milos "baze" Bazelides
;sourced from http://baze.au.com/misc/z80bits.html#4.2

.WhiteNoiseGenerator 
	PUSH	HL		;save registers
	PUSH 	DE

.WhiteNoiseGenStart	
	LD	DE,0		; Seed is usually 0
	LD	A,D
	LD	H,E
	LD	L,253
	OR	A
	SBC	HL,DE
	SBC	A,0
	SBC	HL,DE
	LD	D,0
	SBC	A,D
	LD	E,A
	SBC	HL,DE
	JR	NC,WhiteNoiseGenRand
	INC	HL
.WhiteNoiseGenRand
	LD	(WhiteNoiseGenStart+1),HL ;Self-modifying code

	LD	A,L		;We just want 8-bit value in A
	POP 	DE		;restore registers
	POP	HL
	RET

.PlayWNEnd
	ei			;re-enable interrupts
#endasm
}

void my_print_at(byte row,byte col,char *string,byte attr)
{
char c;
byte pixelrow,i;
byte *src,*dest;

//set global text cursor position
textrow=row;
textcol=col;

while(c=*string++)
	{
	//If attr is non-zero then write attributes
	if(attr)
		*(byte *)(0x5800+(textrow<<5)+textcol)=attr;

	src=(byte *)(FONT_ADDR+((c-32)<<3));	

	pixelrow=textrow<<3;
	dest=(byte *)(0x4000+((pixelrow&0xc0)<<5)+((pixelrow&0x38)<<2)+textcol);
	
	for(i=0;i<8;i++)
		{
		*dest=*src++;
		dest+=0x0100;
		}
	textcol++;
	if(textcol>31)
		{
		textcol=0;
		textrow++;
		}
	}
}

void my_cls(byte attr)
{
memset(0x4000,0,0x1800);

if(attr)
	memset(0x5800,attr,0x0300);
}

//Set the border and clear the screen
void my_borcls(byte attr, byte b)
{
zx_border3(b);
my_cls(attr);
}

void init_anim_lists(void)
{
byte i;
//Anim lists for the player-character Sid.
//Comprise the list of pointers to frames which give the impression
//of walking forwards when animated in this sequence.
//
//Offsets into the array of frames are added onto
//7 bytes padding before the first 48-byte frame
//(Then each frame is 48 bytes apart)
anim_sid_right[0]=sprite_sid1+7;
anim_sid_right[1]=sprite_sid1+55;
anim_sid_right[2]=sprite_sid1+103;
anim_sid_right[3]=sprite_sid1+151;
anim_sid_right[4]=sprite_sid1+103;
anim_sid_right[5]=sprite_sid1+55;

anim_sid_left[0]=sprite_sid1+199;
anim_sid_left[1]=sprite_sid1+247;
anim_sid_left[2]=sprite_sid1+295;
anim_sid_left[3]=sprite_sid1+343;
anim_sid_left[4]=sprite_sid1+295;
anim_sid_left[5]=sprite_sid1+247;

//Set up start frames here too
//These are the frames which are initially drawn when
//the map is drawn
//player sprite
start_frame[0]=sprite_sid1+7;

//guardians
for(i=1;i<NUM_SPRITES;i++)
	start_frame[i]=sprite_guardian1+7+(96*(i-1));
}

//Convert an (unsigned) 16-bit word into a decimal string in "str",
//padded to length 'len' (where len<=6). This function
//is used as sprintf() is not available in this program. (not using stdio.h)
//The wtod() and reverse_string() functions sourced from http://www.jb.man.ac.uk/~slowe/cpp/itoa.html
//and simplified for this program (which needs base 10 only).
//Also added zero-padding 
void wtod(word value, char* str, byte len) 
{
	
char *lstr=str;
byte padding;

do	
	{
	*lstr++='0'+(value%10);
	} while(value/=10);
	
*lstr=0;

//append zero padding to end of string if needed 
padding=len-strlen(str);

//add up to five zeros (one digit already guaranteed to be present in string)
if(padding)
	{
	strncat(str,"00000",padding);
	lstr+=padding;
	}

//String is back to front - so reverse it
reverse_string(str,lstr-1);
}

void reverse_string(char* begin, char* end) 
{
char t;
	
while(end>begin)
	t=*end, *end--=*begin, *begin++=t;
	
}

void add_sprite_colour(struct sp1_cs *cs)
{
//Used to set the colours of the guardian sprites
cs->attr=sprite_colour[global_n];
cs->attr_mask=0;
}

void calculate_sprite_colours(void)
{
byte i;
//Player sprite typically bright yellow
sprite_colour[0]=INK_YELLOW|BRIGHT;

//Guardians
for(i=1;i<NUM_SPRITES;i++)
	{
	//Avoid brick_colour (Sprites must be inited after background in each room)
	//brick colour in range 1-4
	//i in range 1-4
	sprite_colour[i]=(brick_colour&7)+i;
	//check for result higher than 5 (higher than cyan)
	//if so then subtract 5 to bring colour back into needed range
	if(sprite_colour[i]>5)
		sprite_colour[i]-=5;
	//make sprites BRIGHT
	sprite_colour[i]|=BRIGHT;
	}	
}

void draw_status_area(void)
{
draw_room_name();
draw_lives();
draw_nuts();
}

void draw_room_name(void)
{
byte attr=INK_YELLOW|BRIGHT;
//Erase the line
memset(tempstr,' ',32);
tempstr[32]=0;
my_print_at(22,0,tempstr,attr);
wtod(room,tempstr,2);
my_print_at(22,0,"ROOM",attr);
my_print_at(22,5,tempstr,attr);
my_print_at(22,7,":",attr);
my_print_at(22,8,room_name[room-1],attr);
}

void draw_lives(void)
{
wtod(lives,tempstr,2);
my_print_at(23,0,"LIVES:",0);
my_print_at(23,6,tempstr,0);
}

void draw_nuts(void)
{
wtod(nuts,tempstr,3);
my_print_at(23,24,"NUTS:",0);
my_print_at(23,29,tempstr,0);
}

void reset_game_objects(void)
{
byte *obj_ptr,odd;
byte **rp;

//reset nuts count
nuts=0;

//set temporary room pointer to beginning of room list
rp=room_ptr;

//Step through each room's data
while(room_data=*rp++)
	{
	obj_ptr=room_data+O_OFFSET;
	odd=0;
	//step through object list for this room
	while(*obj_ptr!=0xff)
		{
		//check even bytes only (row values)
		if(!odd)
			{
			//reset "collected flag" (high-bit)
			*obj_ptr&=0x7f;
			nuts++;  //add one to the nut count	
			}
		//toggle "odd flag"
		odd^=1;
		obj_ptr++; //next byte
		}
	}
}

void game_over(void)
{
byte i;
my_cls(BRIGHT);
for(i=4;i<20;i++)
	my_print_at(i,4+((i+4)&0x0f),"GAME OVER!",INK_YELLOW|PAPER_RED|BRIGHT|FLASH);
play_white_noise(1000,1,1); //long explosion
WAIT_KEY;
}

void won_game(void)
{
byte i,j;
my_cls(BRIGHT);
for(i=4;i<20;i++)
	my_print_at(i,3+((i+4)&0x07),"YOU COMPLETED SS3!",INK_YELLOW|PAPER_GREEN|BRIGHT|FLASH);
		
	for(j=0;j<12;j++)
		for(i=(j&3)+4;i<12;i++)
			play_tone(100,100-(i<<3),0);
WAIT_KEY;
}

void collect_object(void)
{
//Using globals to save stack space
//this function marks the object in the current room at (row=ty,column=tx)
//as collected by setting the high bit in its row attribute
byte *obj_ptr=room_data+O_OFFSET;
while(*obj_ptr!=0xff)
	{
	if(*obj_ptr==ty)
		if(*(obj_ptr+1)==tx)
			{
			//set high bit in row attribute
			*obj_ptr|=0x80;
			break; //object found - stop searching
			}	
	obj_ptr+=2; //skip odd bytes
	}
}

void lose_a_life(void)
{
	zx_border3(2);
	play_tone(200,1,2);
	zx_border3(0);
	lives--;
	//reset player location and parameters
	se->dx=0;
	se->dy=0;
	airborne=0;
	jump_tick=0;
	//se->endcol and se->endrow in fact refer to initial hrot and vrot values
	//for the player sprite only
	sp1_MoveSprAbs(se->s, &cr, start_frame[0],se->startrow , se->startcol, se->endrow, se->endcol);
	draw_lives();
	//player is given temporary invulnrebility when a life is lost
	invulnerable=INVULNERABLE_TIME;	
}

void play_mrbeep_tune(void)
{

#asm
	call 49152
#endasm

}

